/*****************************************************************************
 * IBM Confidential
 *
 * OCO Source Materials
 *
 * (C) Copyright IBM Corp. 2004 All Rights Reserved.
 *
 * The source code for this program is not published or
 * otherwise divested of its trade secrets, irrespective of
 * what has been deposited with the U.S. Copyright Office.
 *
 * ChangeLog:
 * Sep 24, 2004		Created.
 * Oct 06, 2004		Added check for failure to allocate memory on the heap.
 *                  Corrected some comments.
 * Feb 05, 2007		Modified to Support z/OS
 * Oct 08, 2007		Modified to Correctly mask password on z/OS
 ****************************************************************************/

/**
 * The shared library used to execute native code when it is necessary
 * to do so.
 *
 * Author: Wesley Miaw <wmiaw@us.ibm.com>
 */


/** Flag indicating this is beign compiled for an MVS system.
 * Comment the below line out if compiling for Unix, etc.
 * Uncomment it it compiling for MVS.
 */
#define MVS 1

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "com_ibm_storage_perfcollect_handlers_ConfigHandler.h"

/** Win32 includes. */
#ifdef WIN32
#include <windows.h>

/** Linux includes. */
#else
#include <termios.h>
#include <unistd.h>
#endif

/** MVS includes. */
#ifdef MVS
#include <jni.h>
#include <jni_convert.h>
#include <errno.h>
#include <stropts.h>
#include <sys/ioctl.h>

#endif

/**
 * The maximum ESSNI password length, plus two to check for passwords
 * that are too long by accepting an additional character and the
 * trailing '\0'.
 */
#define MAX_LENGTH 18

/**
 * The format string used to scan in the input. The length of the
 * string should be one less than MAX_LENGTH to accomodate the
 * trailing '\0'.
 */
#define FORMAT "%17s"

/** The return value of readPassword if an error occurs. */
#define EMPTY ""

/** Error string. */
static char *error_string = NULL;

/**
 * Get the password error.
 */
JNIEXPORT jstring JNICALL Java_com_ibm_storage_perfcollect_handlers_ConfigHandler_passwordError (JNIEnv *env, jobject this) {
#ifdef MVS
	jstring retVal;
	/* convert from EBCDIC to Unicode */
	if (error_string == NULL){
		NewStringPlatform(env, "Squadrons_Config.unknownPasswordError", &retVal, 0);
		return retVal;
	}
	else{
		NewStringPlatform(env, error_string, &retVal, 0);
		return retVal;
	}
#else
	if (error_string == NULL) return (*env)->NewStringUTF(env, "Squadrons_Config.unknownPasswordError");
	else return (*env)->NewStringUTF(env, error_string);
#endif
}

/**
 * Set the password error.
 */
void PasswordError(char *error) {
	if (error_string != NULL) free(error_string);
	error_string = (char *)strdup(error);
}

/**
* Duplicates the String and stores it on heap. Refer to posix strdup
*/
char *tpcrstrdup(const char *original){
	size_t length;
	char *retVal;
	length = strlen(original);
	retVal = (char *)malloc(length + 1);

	if(retVal != 0){
		strncpy(retVal, original, length);
		retVal[length] = 0;
	}
	return retVal;
}

/**
 * C-only portion of the readPassword function. The return character
 * array must be freed by the caller.
 *
 * If an error occurs, the empty string is returned and the password
 * error string is set to the error code defined in
 * deviceerrors.properties. If memory could not be allocated on the
 * heap, NULL is returned instead of the empty string.
 */
char* ReadPassword() {
	/**
	 * The input buffer.
	 */
	char buffer[MAX_LENGTH];
	int response;

	/* Old terminal settings. */
#ifdef WIN32
	HANDLE in;
	BOOL rc;
	DWORD mode;
#else
	struct termios oldtty, tty;
	int rc;
#endif
#ifdef MVS
	int beginInt 	= 0x01000000;
	int endInt 	= 0x02000000;
	int retVal 		= 0;
#endif

	/* Clear the buffer. */
	memset(buffer, 0, (MAX_LENGTH) * sizeof(char));

	/* Turn off terminal echo, saving the previous terminal settings. */
#ifdef WIN32
	in = GetStdHandle(STD_INPUT_HANDLE);
	rc = GetConsoleMode(in, &mode);
	if (rc == 0) {
		PasswordError("Squadrons_Config.GetConsoleMode");
		return strdup(EMPTY);
	}
	rc = SetConsoleMode(in, (mode | ENABLE_LINE_INPUT) & ~ENABLE_ECHO_INPUT);
	if (rc == 0) {
		PasswordError("Squadrons_Config.SetConsoleMode");
		return strdup(EMPTY);
	}
#else
	rc = tcgetattr(0, &oldtty);
	if (rc != 0) {
		PasswordError("Squadrons_Config.tcgetattr");
		return (char *)strdup(EMPTY);
	}
	tty = oldtty;
	tty.c_lflag &= ~ECHO;
	rc = tcsetattr(0, TCSANOW, &tty);
	if (rc != 0) {
		PasswordError("Squadrons_Config.tcsetattr");
		return (char *)strdup(EMPTY);
	}
#endif
#ifdef MVS
	/* Turning on masking code in c code */
	retVal = ioctl(STDIN_FILENO, TIOCNOTIFY, &beginInt);
	if(retVal != 0)
	{
		printf("Error hiding STDIN: errno is %d\n", errno);
		perror("Error hiding STDIN");
	}
#endif

	/**
	 * Read in the user's password.
	 *
	 * After doing so, flush standard input so extra characters are
	 * gone. Cannot use fflush(stdin) because it is undefined for
	 * standard in.
	 */
	response = scanf(FORMAT, buffer);

	if (response != 1) {
		PasswordError("Squadrons_Config.scanf");
		return (char *)strdup(EMPTY);
	}

	while (getchar() != '\n');

	/* Restore the terminal settings. */
#ifdef WIN32
	SetConsoleMode(in, mode);
#else
	tcsetattr(0, TCSANOW, &oldtty);
#endif
#ifdef MVS
	/* Turning off masking in c code\ */
	retVal = ioctl(STDIN_FILENO, TIOCNOTIFY, &endInt);
	if(retVal != 0)
	{
		printf("Error revealing STDIN: errno is %d\n", errno);
		perror("Error revealing STDIN");
	}
#endif

	/* Check the password. */
	if (strcmp(buffer, EMPTY) == 0) {
		PasswordError("Squadrons_Config.emptyPassword");
		return (char *)strdup(EMPTY);
	}
	/* Return the password. */
	/*return (char *)strdup(buffer);*/
	return tpcrstrdup(buffer);
}

/**
 * Prompts the user for a password with terminal echo turned off.
 *
 * Returns the empty string if there was an error.
 */
JNIEXPORT jstring JNICALL Java_com_ibm_storage_perfcollect_handlers_ConfigHandler_readPassword (JNIEnv *env, jobject this) {
	/* Read the password. */
	jstring j_password;
	char *password = ReadPassword();
	/**
	 * If NULL was returned, memory could not be allocated on the heap.
	 * Don't try setting the error string, as that requires memory to be
	 * allocated on the heap.
	 */
#ifdef MVS
	/* convert from EBCDIC to Unicode */
	if (password != NULL) {
		NewStringPlatform(env, (const char *)password, &j_password, 0);
	}
	else{
		NewStringPlatform(env, (const char *)EMPTY, &j_password, 0);

	}
#else
	if (password != NULL) {
		j_password = (*env)->NewStringUTF(env, password);
	} else {
		j_password = (*env)->NewStringUTF(env, EMPTY);
	}
#endif
	/* Clean up and return the password. */
	free(password);
	return j_password;
}
